import * as http from 'http';
import * as ioserver from 'socket.io';
import * as ioclient from 'socket.io-client';

export type ReturnFunction = (returnedValue: any, thrownException: any) => any;
export type CodeFunction = (returnMethod: ReturnFunction, args: any[]) => void;
export type EventFunction = (...args: any[]) => void;

class DeclaredMethod {
  name: string;
  code: CodeFunction;
  constructor(name: string, code: CodeFunction) {
    this.name = name;
    this.code = code;
  }
}
class RemoteError extends Error {
  constructor(thrownException: any) {
    super(thrownException.message);
    this.name = thrownException.remote_name;
    this.message = thrownException.remote_message;
    this.stack = thrownException.remote_stack;
  }
}

interface CookieAndArgs {
  cookie: string;
  args: any[];
}
interface ReturnedValueAndThrownException {
  returnedValue: any;
  thrownException: any;
}

function S4(): string {
  // tslint:disable-next-line no-bitwise
  return (((1 + Math.random()) * 0x10000) | 0).toString(16).substring(1);
}
function GUID(): string {
  return `${S4()}${S4()}-${S4()}-4${S4().substr(0,3)}-${S4()}-${S4()}${S4()}${S4()}`;
}

export class SioRpcServer {

  io: SocketIO.Server;
  connectedSockets: SocketIO.Socket[];
  private declaredMethods: DeclaredMethod[];

  constructor(httpServer: http.Server) {
    this.io = ioserver(httpServer);
    this.connectedSockets = [];
    this.declaredMethods = [];
  }
  declare(methodName: string, methodCode: Function) {
    this.declaredMethods.push(
      new DeclaredMethod(
        methodName,
        async (returnFunction, args) => {
          let result: any;
          try {
            result = await methodCode.apply(null, args);
          } catch (exception) {
            return returnFunction(undefined, exception);
          }
          return returnFunction(result, undefined);
        }
      )
    );
  }
  publish(eventName: string, ...args: any[]) {
    for (const connectedSocket of this.connectedSockets) {
      connectedSocket.emit(`${eventName}..event`, args);
    }
  }
  run() {
    this.io.on('connection', (socket) => {
      // Remember this socket
      this.connectedSockets.push(socket);
      socket.on('disconnect', () => {
        // Forget this socket
        const socketIndex = this.connectedSockets.indexOf(socket);
        if (socketIndex !== -1) {
          this.connectedSockets.splice(socketIndex, 1);
        }
      });
      // Make all declared methods available on this socket
      for (const declaredMethod of this.declaredMethods) {
        socket.on(`${declaredMethod.name}..call`, (cookieAndArgs: CookieAndArgs) => {
          const returnMethod = (value: any, exception: Error) => {
            const returnedValue = value;
            let thrownException: any;
            if (typeof exception !== 'undefined') {
              thrownException = {
                remote_name: exception.name,
                remote_message: exception.message,
                remote_stack: exception.stack,
              };
            }
            const returnedValueAndThrownException = <ReturnedValueAndThrownException> { returnedValue, thrownException };
            socket.emit(`${declaredMethod.name}..return..${cookieAndArgs.cookie}`, returnedValueAndThrownException);
          };
          declaredMethod.code(returnMethod, cookieAndArgs.args);
        });
      }
    });
  }

}

export class SioRpcClient {

  socket: SocketIOClient.Socket;

  constructor(url: string) {
    this.socket = ioclient.connect(url);
  }
  call(methodName: string, ...args: any[]): Promise<any> {
    return new Promise((resolve, reject) => {
      const cookie = GUID();
      this.socket.once(`${methodName}..return..${cookie}`, (returnedValueAndThrownException: ReturnedValueAndThrownException) => {
        if (typeof returnedValueAndThrownException.thrownException !== 'undefined') {
          return reject(new RemoteError(returnedValueAndThrownException.thrownException));
        }
        resolve(returnedValueAndThrownException.returnedValue);
      });
      this.socket.emit(`${methodName}..call`, <CookieAndArgs>{ cookie, args });
    });
  }
  subscribe(eventName: string, eventCode: EventFunction) {
    this.socket.on(`${eventName}..event`, (args: any[]) => {
      eventCode.apply(null, args);
    });
  }
  subscribeOnce(eventName: string, eventCode: EventFunction) {
    this.socket.once(`${eventName}..event`, (args: any[]) => {
      eventCode.apply(null, args);
    });
  }

}
